#-----------------------------------------------------------------------------
# Name        :   Bezier Triangle Tool 2.0
# Description :   A tool to create and edit rational Bezier triangles.
# Menu Item   :   Draw->Bezier Triangle
# Context Menu:   Edit Bezier Triangle
# Usage       :   Draw->Bezier Triangle will draw a default rational Bezier 
#                 triangle and put you into the editing tool for the newly 
#                 drawn triangle.
#             :   Selecting an existing rational Bezier triangle drawn/edited
#                 with this tool and using Edit Bezier Triangle in the Right 
#                 Mouse Button context menu will also activate the editing tool
#                 for the selected triangle.
#             :   To change the weight of a non-end control point, hover the
#                 cursor over the control point you want to change the weight
#                 of so that it is highlighted in green, type the new weight
#                 you want for it, without moving the cursor away from the
#                 control point, and press enter or return to influence the 
#                 shape of the triangle.
#             :   The editing tool also allows you to select either the 10 
#                 control points or the 18 lines of the control polygon that 
#                 can be moved to manipulate the rational Bezier triangle to be
#                 the way you want it to be.
#                 Inferencing is available when the selected points are moved.
#                 Also, inference locking is available using the Shift key.
#                 If the user holds down the shift key for more than 1/2 second,
#                 then the infrences are unlocked on the release.  Otherwise, 
#                 the user presses shift once to lock and a second time to 
#                 unlock.
# Date        :   01/24/2015
# Type        :   Tool
#-----------------------------------------------------------------------------

# First we pull in the standard API hooks.
require 'sketchup.rb'

#-----------------------------------------------------------------------------
module JPS::BezierTriangle

#--------------------------------------------------------------------------------------------------------------------------------
class JPS_BezTri

  def initialize(ctrlPoints, weights, nstrips)
    @ctrlPoints = ctrlPoints
    @weights = weights
    @nstrips = nstrips
    @n = ((-3 + Math.sqrt(1 + 8*ctrlPoints.length))/2).to_i
    @n_fact = factorial(@n)
    @surfpnts = Array.new(((@nstrips+1)*(@nstrips+2))/2)
    generateSurfacePoints()
  end

  def factorial( n )
    fact = 1
    
    for i in 2..n
      fact *= i
    end
    
    return fact
  end
  
  def evalCoord(u, v)
    surfpnt = [0.0, 0.0, 0.0]
    den = 0.0
    i_pnt = 0
    for  j in 0 .. @n 
      j_fact = factorial(j)
      for i in 0 .. @n - j
        i_fact = factorial(i)
        k = @n - i -j;
        k_fact = factorial(k)
        bern = (@n_fact/(i_fact*j_fact*k_fact))*(u**i)*(v**j)*((1.0-u-v)**k)
        den += @weights[i_pnt]*bern
        for coord in 0 .. 2
          surfpnt[coord] += @weights[i_pnt]*bern*@ctrlPoints[i_pnt][coord]
        end
        i_pnt += 1
      end
    end
    return Geom::Point3d.new(surfpnt[0]/den, surfpnt[1]/den, surfpnt[2]/den)
  end

  def generateSurfacePoints()
    i_pnt = 0
    for j in 0..@nstrips
      for i in 0..@nstrips-j
        @surfpnts[i_pnt] = evalCoord(i.to_f/@nstrips, j.to_f/@nstrips)
        i_pnt += 1
      end
    end
  end

  def addTrianglesToModel(group)
    
    my_mesh = Geom::PolygonMesh.new

    i_pnt = -2
    i_pnt_on_next_line = @nstrips
    for j in 1..(@nstrips)
      i_pnt += 2
      i_pnt_on_next_line += 1
      my_mesh.add_polygon @surfpnts[i_pnt], @surfpnts[i_pnt+1], @surfpnts[i_pnt_on_next_line] 
      for i in 2..(@nstrips-j+1)
        i_pnt += 1
        my_mesh.add_polygon @surfpnts[i_pnt], @surfpnts[i_pnt_on_next_line+1], @surfpnts[i_pnt_on_next_line] 
        my_mesh.add_polygon @surfpnts[i_pnt], @surfpnts[i_pnt+1], @surfpnts[i_pnt_on_next_line+1] 
        i_pnt_on_next_line += 1
      end
    end

    model = Sketchup.active_model
 
    model.start_operation "Bezier Triangle", true

      if(group == nil)
        group = model.active_entities.add_group
      else
        group.entities.clear!
        group.transformation = Geom::Transformation.new
      end

      group.entities.fill_from_mesh my_mesh

      group.set_attribute("jps", "surftype", "Rational BezierTriangle")
      group.set_attribute("jps", "surfctrlpnts", @ctrlPoints)
      group.set_attribute("jps", "weights", @weights)
  
    model.commit_operation
 
    return group
  end
  
  def display(view)

    view.drawing_color = Sketchup::Color.new "Blue"

    k = 0
    for i in 0..@nstrips-1
      polyline = Array.new(@nstrips+1-i)
      for j in 0..@nstrips-i
        polyline[j] = @surfpnts[k]
        k += 1
      end
      view.draw(GL_LINE_STRIP, polyline)
    end

    for i in 0..@nstrips-1
      polyline = Array.new(@nstrips+1-i)
      for j in 0..@nstrips-i
        polyline[j] = @surfpnts[i+(@nstrips+1)*j-(j*(j-1))/2]
      end
      view.draw(GL_LINE_STRIP, polyline)
    end
    
    for i in 0..@nstrips-1
      polyline = Array.new(i+2)
      for j in 0..i+1
        polyline[j] = @surfpnts[i+1+@nstrips*j-(j*(j-1))/2]
      end
      view.draw(GL_LINE_STRIP, polyline)
    end

  end

end

#--------------------------------------------------------------------------------------------------------------------------------

# Function to test to see if the selection set contains only a Bezier Triangle
# Returns the triangle if there is one or else nil
def self.jps_selected_bezier_triangle
    ss = Sketchup.active_model.selection
    return nil if ss.nitems != 1
    group = ss.first
    return nil if not group.kind_of? Sketchup::Group
    return nil if (group.get_attribute("jps", "surftype") != "BezierTriangle" &&
                   group.get_attribute("jps", "surftype") != "Rational BezierTriangle")
    return group
end

#------------------------------------------------------------------------------------------------------------------------------------
# Create the default Bezier triangle

def self.jps_draw_bezier_triangle
  
  ctrlPoints = [ [-30, -30, 20], [-10, -30, 70], [10, -30, 70], [30, -30, 20],
                          [-20, -10, 70], [ 0, -10, 20], [20, -10, 70],
                                 [-10,  10, 70], [10,  10, 70],
                                          [0,  30, 20] ]
 
  
  weights = [1.0, 2.0, 2.0, 1.0,
               2.0, 0.5, 2.0,
                 2.0, 2.0,
                   1.0]
 
  bez_tri = JPS_BezTri.new(ctrlPoints, weights, 10)
  return (bez_tri.addTrianglesToModel(nil))
  
end

#-----------------------------------------------------------------------------
# Define the tool class for editing Bezier triangles

class JPS_EditBezierTriangleTool
  
  def generateControlLines()
    for i in 0..3*@n-1
      for j in 0..@n-1-(i%@n)
        @ctrlLines[i][j] = @ctrlPoints[@map_ctrlLines_to_ctrlPoints[i][j][0]]
      end
      @ctrlLines[i][@n-(i%@n)] = @ctrlPoints[@map_ctrlLines_to_ctrlPoints[i][j][1]]
    end
  end

  # Reset the tool back to its initial state
  def reset(view)
    if (@state == 1 && @selection)    
      if( @selection >= 0 ) 
        if (@selection%(@n+1) == @ctrlLines[@selection/(@n+1)].length - 1)
          @ctrlPoints[@map_ctrlLines_to_ctrlPoints[@selection/(@n+1)][(@selection-1)%(@n+1)][1]] = @pt_to_move.position.to_a
        else  
          @ctrlPoints[@map_ctrlLines_to_ctrlPoints[@selection/(@n+1)][@selection%(@n+1)][0]] = @pt_to_move.position.to_a
        end
      else
        @ctrlPoints[@map_ctrlLines_to_ctrlPoints[(-@selection-1)/(@n+1)][(-@selection-1)%(@n+1)][0]] = @segment[0].to_a
        @ctrlPoints[@map_ctrlLines_to_ctrlPoints[(-@selection-1)/(@n+1)][(-@selection-1)%(@n+1)][1]] = @segment[1].to_a
      end    
    end
    generateControlLines()
    @selection = nil
    @state = 0
    @pt_to_move = nil
    # clear the InputPoint
    @ip.clear
    if( view )
        view.tooltip = nil
        view.invalidate if @drawn
    end
    @drawn = false
    @user_did_undo = false
  end  
  
  def initialize_map_ctrlLines_to_ctrlPoints()
    i_pnt = 0
    for i in 0..@n-1
      for j in 0..@n-1-i
        @map_ctrlLines_to_ctrlPoints[i][j][0] = i_pnt
        @map_ctrlLines_to_ctrlPoints[i][j][1] = i_pnt+1
        i_pnt += 1
      end
      i_pnt += 1
    end
    
    for i in @n..2*@n-1
      i_pnt = i-@n
      for j in 0..2*@n-1-i
        @map_ctrlLines_to_ctrlPoints[i][j][0] = i_pnt
        @map_ctrlLines_to_ctrlPoints[i][j][1] = i_pnt+@n+1-j
        i_pnt += @n+1-j
      end
    end

    for i in 2*@n..3*@n-1
      i_pnt = 3*@n-i
      for j in 0..@n-1-(i%@n)
        @map_ctrlLines_to_ctrlPoints[i][j][0] = i_pnt
        @map_ctrlLines_to_ctrlPoints[i][j][1] = i_pnt+@n-j
        i_pnt += @n-j
      end
    end
  end

  def initialize(group)
    @group = group
    @ctrlPoints = group.get_attribute("jps", "surfctrlpnts")
    @n_ctrlPoints = @ctrlPoints.length
    transform = group.transformation
    if (not transform.identity? )
      for i in 0..@n_ctrlPoints-1
        @ctrlPoints[i] = transform * @ctrlPoints[i]
      end
    end
    @weights = group.get_attribute("jps", "weights")
    if( not @weights ) 
      @weights = Array.new(@n_ctrlPoints){ 1.0 }
    end
    @n = ((-3 + Math.sqrt(1 + 8*@n_ctrlPoints))/2).to_i
    @ctrlLines = Array.new(3*@n)
    @map_ctrlLines_to_ctrlPoints = Array.new(3*@n)
    for i in 0..3*@n-1
      @ctrlLines[i] = Array.new(@n+1-(i%@n))
      @map_ctrlLines_to_ctrlPoints[i] = Array.new(@n-(i%@n)){ Array.new(2) }
    end
    initialize_map_ctrlLines_to_ctrlPoints()
    @ctrlPointselected = -1

    # @ip is a temporary input point used to get other positions
    @ip = Sketchup::InputPoint.new
    reset(nil)
  end

  def activate
    Sketchup.status_text = "Hover over a control point, type the weight for it and press enter or return or move the control points/lines to influence the shape of the surface"
  end

  # For this tool, allow vcb text entry while the tool is active.
  def enableVCB?
    return true
  end

  def validate_newweight(newweight)
    for i in 0..3*@n-1
      for j in 0..@map_ctrlLines_to_ctrlPoints[i].length-1
        if (@map_ctrlLines_to_ctrlPoints[i][j][0] == @ctrlPointselected)
          if (newweight < 0.1*@weights[@map_ctrlLines_to_ctrlPoints[i][j][1]])
            return -1
          elsif (newweight > 10.0*@weights[@map_ctrlLines_to_ctrlPoints[i][j][1]])
            return 1
          end
        end
        if (@map_ctrlLines_to_ctrlPoints[i][j][1] == @ctrlPointselected)
          if (newweight < 0.1*@weights[@map_ctrlLines_to_ctrlPoints[i][j][0]])
            return -1
          elsif (newweight > 10.0*@weights[@map_ctrlLines_to_ctrlPoints[i][j][0]])
            return 1
          end
        end
      end
    end
    return 0
  end

  def onUserText(text, view)
    if (@ctrlPointselected)
      # get the weight from the text
      newweight = text.to_f
      if(@ctrlPointselected == 0 || @ctrlPointselected == @n || @ctrlPointselected == @n_ctrlPoints-1)
        UI.messagebox "Lets leave the weight of the end control points at 1.0 in standard form."
      elsif(newweight < 0.1)
        UI.messagebox "Minimum weight is 0.1 for numerical stability.\nPlease consider increasing the weights of other control points instead." 
        Sketchup::set_status_text @weights[@ctrlPointselected], SB_VCB_VALUE
      elsif(newweight > 10.0)
        UI.messagebox "Maximimum weight is 10.0 for numerical stability.\nPlease consider decreasing the weights of other control points instead." 
        Sketchup::set_status_text @weights[@ctrlPointselected], SB_VCB_VALUE
      elsif((newweight_ok = validate_newweight(newweight)) == -1)
        UI.messagebox "Minimum weight is one-tenth of adjacent weights for numeric stability.\nPlease consider decreasing the weight(s) of the offending adjacent control point(s) first."
        Sketchup::set_status_text @weights[@ctrlPointselected], SB_VCB_VALUE
      elsif(newweight_ok == 1)
        UI.messagebox "Maximimum weight is ten times adjacent weights for numeric stability.\nPlease consider increasing the weight(s) of the offending adjacent control point(s) first."
        Sketchup::set_status_text @weights[@ctrlPointselected], SB_VCB_VALUE
      else
        @weights[@ctrlPointselected] = newweight
        # Update the actual surface.  
        bez_surf = JPS_BezTri.new(@ctrlPoints, @weights, 10)
        bez_surf.addTrianglesToModel(@group)
      end
    else 
      UI.messagebox "Hover over the control point you want to change the weight of."
      Sketchup::set_status_text "", SB_VCB_VALUE
    end
  end
  
  def deactivate(view)
    view.invalidate if @drawn
    @ip = nil
  end
    
  def resume(view)
    @drawn = false
  end

  def pick_point_to_move(x, y, view)
    old_pt_to_move = @pt_to_move
    @pt_to_move = nil
    ph = view.pick_helper
    @ctrlPointselected = -1
    pickray = view.pickray x, y
    
    for i in 0..3*@n-1
      if( @state == 0 )
        @selection = ph.pick_segment @ctrlLines[i], x, y, 10
        if ( @selection )
          if( @selection < 0 )
            # We got a point on a segment.
            # Compute the point closest to the pick ray.
            @segment = [Geom::Point3d.new(@ctrlLines[i][-@selection-1]), Geom::Point3d.new(@ctrlLines[i][-@selection])]
            result = Geom.closest_points @segment, pickray
            @pt_to_move = Sketchup::InputPoint.new(result[0])
            @selection -= i*(@n+1)
          else
            # we got a control point
            @pt_to_move = Sketchup::InputPoint.new(@ctrlLines[i][@selection])
            if (@selection == @ctrlLines[i].length - 1)
              @ctrlPointselected = @map_ctrlLines_to_ctrlPoints[i][@selection-1][1]
            else  
              @ctrlPointselected = @map_ctrlLines_to_ctrlPoints[i][@selection][0]
            end
            @selection += i*(@n+1)
          end
          @state = 1
        end
      end       
    end

    not (old_pt_to_move == @pt_to_move)
  end  
    
  def onLButtonDown(flags, x, y, view)

    # Select the segment or control point to move
    self.pick_point_to_move x, y, view
    
    # Clear any inference lock
    view.lock_inference
  end
  
  def onLButtonUp(flags, x, y, view)
    return if @state == 0
    @state = 0
    @pt_to_move = nil
    # Update the actual surface.  
    bez_tri = JPS_BezTri.new(@ctrlPoints, @weights, 10)
    bez_tri.addTrianglesToModel(@group)
  end
  
  def onMouseMove(flags, x, y, view)
    # Make sure that the control polygon is shown
    view.invalidate if not @drawn
    
    if (@state == 0) # See if we can select something to move
      if( self.pick_point_to_move(x, y, view) )    
        view.invalidate
        if (@ctrlPointselected > 0 && @ctrlPointselected != @n && @ctrlPointselected != (@n_ctrlPoints-1))
          Sketchup::set_status_text "Weight", SB_VCB_LABEL
          Sketchup::set_status_text @weights[@ctrlPointselected], SB_VCB_VALUE
        else
          Sketchup::set_status_text "", SB_VCB_LABEL
          Sketchup::set_status_text "", SB_VCB_VALUE
        end
      end
      @state = 0
    elsif (@selection) # Move the selected point
      @ip.pick view, x, y, @pt_to_move
      return if not @ip.valid?
      
      if( @selection >= 0 )
        # Moving a control point 
        @ctrlPoints[@ctrlPointselected] = @ip.position.to_a
      else
        # moving a segment
        vec = @ip.position - @pt_to_move.position
        @ctrlPoints[@map_ctrlLines_to_ctrlPoints[(-@selection-1)/(@n+1)][(-@selection-1)%(@n+1)][0]] = (@segment[0].offset vec).to_a
        @ctrlPoints[@map_ctrlLines_to_ctrlPoints[(-@selection-1)/(@n+1)][(-@selection-1)%(@n+1)][1]] = (@segment[1].offset vec).to_a
      end      
      generateControlLines()
      view.tooltip = @ip.tooltip
      view.invalidate
    end
  end  

  # onKeyDown is called when the user presses a key on the keyboard.
  # We are checking it here to see if the user pressed the shift key
  # so that we can do inference locking
  def onKeyDown(key, repeat, flags, view)
    if( key == CONSTRAIN_MODIFIER_KEY && repeat == 1 )
        @shift_down_time = Time.now
        
        # if we already have an inference lock, then unlock it
        if( view.inference_locked? )
            # calling lock_inference with no arguments actually unlocks
            view.lock_inference
        elsif( @state > 0 && @ip.valid? )
            view.lock_inference @pt_to_move, @ip
        elsif( @state > 0 && @pt_to_move.valid? )
            view.lock_inference @pt_to_move
        end
    end
  end

  # onKeyUp is called when the user releases the key
  # We use this to unlock the inference
  # If the user holds down the shift key for more than 1/2 second, then we
  # unlock the inference on the release.  Otherwise, the user presses shift
  # once to lock and a second time to unlock.
  def onKeyUp(key, repeat, flags, view)
    if( key == CONSTRAIN_MODIFIER_KEY &&
        view.inference_locked? &&
        (Time.now - @shift_down_time) > 0.5 )
        view.lock_inference
    end
  end
  
  def getMenu(menu)
    menu.add_item("Done") {Sketchup.active_model.select_tool nil}
  end

  def getExtents
    if( @state > 0 )
      # We are moving a point
      bb = Geom::BoundingBox.new
      if( @ip.valid? )
        bb.add @ip.position
      end
      bb
    end
  end
 
  def drawControlPolygon(view)
    for i in 0..3*@n-1
      view.draw(GL_LINE_STRIP, @ctrlLines[i])
    end
  end

  def drawControlPoints(view)
    for i in 0..@n_ctrlPoints-1
      view.draw_text(view.screen_coords(@ctrlPoints[i]), sprintf("%.2f", @weights[i]))
      view.draw_points(@ctrlPoints[i], 10, 1, "green")
    end
  end     

  # The draw method is called whenever the view is refreshed.  It lets the
  # tool draw any temporary geometry that it needs to.
  def draw(view)

    if (@user_did_undo)
      group = JPS::BezierTriangle::jps_selected_bezier_triangle
      if (group == nil)
        @n = @n_ctrlPoints = @selection = -1
        Sketchup.active_model.select_tool nil
      else
        initialize(group)
      end
      @user_did_undo = false
    end
    
    # Draw the control polygon
    drawControlPolygon(view)
   
    if( @state == 0 )
      if( @pt_to_move )
        view.draw_points(@pt_to_move.position, 10, 1, "red");
      end
    elsif (@selection)
      # Draw the new surface if in "preview"
      bez_tri = JPS_BezTri.new(@ctrlPoints, @weights, 10)
      bez_tri.display(view)
        
      #Draw the lock direction inference clue (the heavy,by-axis color) 
      if( @ip.valid? ) 
        @pt_to_move.draw(view)
        @ip.draw(view)
            
        # The set_color_from_line method determines what color
        # to use to draw a line based on its direction.  For example
        # red, green or blue.
        view.set_color_from_line(@pt_to_move, @ip)
        view.draw_line(@pt_to_move.position, @ip.position)
      end
    end    

    # Draw the control points
    drawControlPoints(view)
    
    if(@ctrlPointselected >= 0)
      view.draw_points(@ctrlPoints[@ctrlPointselected], 10, 2, "green")
    end
    
    @drawn = true
  end

  # The reason can be one of the following values: 
  # 0: the user canceled the current operation by hitting the escape key.
  # 1: the user re-selected the same tool from the toolbar or menu.
  # 2: the user did an undo while the tool was active.
  def onCancel(reason, view)
    self.reset(view)
    if(reason == 2)
      @user_did_undo = true 
    end
  end

end # class EditBezierTriangleTool


#-----------------------------------------------------------------------------
# First check to see if we have already loaded this file so that we only 
# add the item to the menu once
if( not file_loaded?("jps_beziertriangle.rb") )

  # Show the Ruby Console at startup so we can
  # see any programming errors we may make.
  # Sketchup.send_action "showRubyPanel:"

  # Add a menu choice for creating bezier triangles
  UI.menu("Draw").add_item("Bezier Triangle") {
    model = Sketchup.active_model
    selection = model.selection
    selection.clear
    group = self.jps_draw_bezier_triangle
    #selection.add group
    model.select_tool JPS_EditBezierTriangleTool.new(group)
  }

  # Add a context menu handler to let you edit a Bezier triangle
  UI.add_context_menu_handler do |menu|
    if( (group = self.jps_selected_bezier_triangle) )
      item = menu.add_item("Edit Bezier Triangle") {
        Sketchup.active_model.selection.clear
        Sketchup.active_model.select_tool JPS_EditBezierTriangleTool.new(group)
      }
      menu.set_validation_proc(item) {
        if group.locked?
          MF_GRAYED
        end
      }
    end
  end
    
  # Let Ruby know we have loaded this file
  file_loaded("jps_beziertriangle.rb")

end

end #module JPS::BezierTriangle